#include "rpi_lib/rpi.h"
#include <stdio.h>
#include <stdint.h>

// 4x + y + z = 9
// x + 3y + z = 10
// 2x + y + 5z = 19

#include "Ab_192.h"

double c[SIZE];
double x[SIZE];

extern void enable_fpu(void);

void myLUsolver(void){

	int i,j,k;
	int n = SIZE;

	// 外積方式ガウス法

	// LU分解
	for (k=0; k<n; k++) {
		// kで回す行方向ループ

		// 行列Lの処理
		double div_tmp = 1.0 / A[k][k];
		for (i=k+1; i<n-12; i+=12) {
			// iで回す行方向ループ

			double d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;

			// l_ik = a_ik / a_kk
			d4 = A[i][k];
			d5 = A[i+1][k];
			d6 = A[i+2][k];
			d7 = A[i+3][k];
			d8 = A[i+4][k];
			d9 = A[i+5][k];
			d10 = A[i+6][k];
			d11 = A[i+7][k];
			d12 = A[i+8][k];
			d13 = A[i+9][k];
			d14 = A[i+10][k];
			d15 = A[i+11][k];

			d4 *= div_tmp;			
			d5 *= div_tmp;
			d6 *= div_tmp;
			d7 *= div_tmp;
			d8 *= div_tmp;
			d9 *= div_tmp;
			d10 *= div_tmp;
			d11 *= div_tmp;
			d12 *= div_tmp;
			d13 *= div_tmp;
			d14 *= div_tmp;
			d15 *= div_tmp;

			A[i][k] = d4;
			A[i+1][k] = d5;
			A[i+2][k] = d6;
			A[i+3][k] = d7;
			A[i+4][k] = d8;
			A[i+5][k] = d9;
			A[i+6][k] = d10;
			A[i+7][k] = d11;
			A[i+8][k] = d12;
			A[i+9][k] = d13;
			A[i+10][k] = d14;
			A[i+11][k] = d15;
		}
		for(;i<n;i++){
			A[i][k] = A[i][k] * div_tmp;
		}

		// 行列Lの処理
	// 	double div_tmp = 1.0 / A[k][k];
	// 	for (i=k+1; i<n; i++) {
	// 		// iで回す行方向ループ

	// 		// l_ik = a_ik / a_kk
	// 		A[i][k] = A[i][k] * div_tmp;   
	// 	}


		// 行列Uの処理
		for (j=k+1; j<n; j++) {
			// jで回す行方向のループ

			// Tips: ここで値を変数に入れたほうが
			// ループ内で参照するよりメモリアクセス回数が減ってお得
			double a_jk = A[j][k];

			// for (i=k+1; i<n; i++) {
			// 	// iで回す列方向のループ
			// 	A[j][i] = A[j][i] - A[k][i] * a_jk;
			// }

			for (i=k+1; i<n-6; i+=6) {
				// iで回す列方向ループ
				double d4,d5,d6,d7,d8,d9;

				// l_ik = a_ik / a_kk
				d4 = A[k][i+0];
				d5 = A[k][i+1];
				d6 = A[k][i+2];
				d7 = A[k][i+3];
				d8 = A[k][i+4];
				d9 = A[k][i+5];

				d4 *= a_jk;			
				d5 *= a_jk;
				d6 *= a_jk;
				d7 *= a_jk;
				d8 *= a_jk;
				d9 *= a_jk;

				A[j][i+0] -= d4;
				A[j][i+1] -= d5;
				A[j][i+2] -= d6;
				A[j][i+3] -= d7;
				A[j][i+4] -= d8;
				A[j][i+5] -= d9;
			}
			for(;i<n;i++){
				A[j][i] = A[j][i] - A[k][i] * a_jk;
			}
		}


	}
	// LU分解終了


	// 前進代入
	// Lc = b
	for(k=0; k<n; k++){
		// 行方向のループ
		c[k] = b[k];
		for(i=0; i<k; i++){
			// 列方向のループ
			c[k] -= A[k][i] * c[i];
		}
	}


	// 後進代入
	// Ux = c
	// k = n - 1
	x[n-1] = c[n-1] / A[n-1][n-1];
	// k = (n - 2) ~ (0)
	for(k = n - 2; k >= 0; k--){
		x[k] = c[k];
		for(i=k+1; i<n; i++){
			// 列方向のループ
			x[k] -= A[k][i] * x[i];
		}
		x[k] = x[k] / A[k][k];
	}
}


int main(void){
	rpi_init();

	enable_fpu();


	int i,j,k;
	int n = SIZE;

	for(i = 0;i<100;i++){
		printf(" ");
	}
	printf("\n");


	uint64_t start_time = get_systime();

	myLUsolver();

	uint64_t end_time = get_systime();

	// xを出力
	printf("x:\n");
	for(i=0;i<n;i++){
		printf("% 2.2f ",x[i]);
	}
	printf("\n");


	char str[256];

	sprintf(str,"start time: %lld\n",start_time);
	printf("%s",str);
	sprintf(str,"end time: %lld\n",end_time);
	printf("%s",str);
	sprintf(str,"exec time: %lld\n",end_time - start_time);
	printf("%s",str);

	return 0;
}
